#!/usr/local/bin/perl -w
#
# SLB - Super Light Blog
#
# Copyright (c) 2005,2006 Satoshi Fukutomi <info@fuktommy.com>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#

use strict;
use Carp;
use CGI;
use lib ("../lib");

my $config;

package Error; #{
    sub error($) {
        my ($msg) = @_;
        print "Content-Type: application/xml\n";
        print "\n";
        print '<?xml version="1.0" encoding="utf-8"?>', "\n";
        print sprintf('<error>%s</error>', $msg), "\n";
    }

    sub xdie($) {
        my ($msg) = @_;
        error($msg);
        exit;
    }
#}

package BlogConfig; #{
    sub new($$) {
        my ($self, $filename) = @_;
        $self = bless {};
        $self->read($filename);
        return $self;
    }

    sub read($$) {
        my ($self, $filename) = @_;
        my $fp;
        open $fp, $filename;
        local $_;
        while (<$fp>) {
            my ($key, $val) = split /:/, $_, 2;
            $key =~ s/^\s*|\s*$//g;
            $val =~ s/^\s*|\s*$//g;
            $self->{$key} = $val;
        }
        close $fp;
    }
#}

package Item; #{
    sub new($%) {
        my ($self, %opts) = @_;
        return bless {%opts};
    }
#}

package RSS; #{
    my $title = "";
    my $parent = "";        # Place where is documents or RSS
    my $link = "";          # URI of main page
    my $uri = "";           # URI of RSS
    my $description = "";

    sub new($%) {
        my ($self, %opts) = @_;
        $self = {};
        $self->{encode} = "utf-8";
        $self->{lang} = "ja";
        $self->{items} = {};

        if ($opts{title}) {
            $self->{title} = $opts{title};
        } else {
            $self->{title} = $config->{title};
        }
        if ($opts{parent}) {
            $self->{parent} = $opts{parent};
        } else {
            $self->{parent} = $config->{link};
        }
        if ($opts{link}) {
            $self->{link} = $opts{link};
        } else {
            $self->{link} = $config->{link};
        }
        if ($opts{uri}) {
            $self->{uri} = $opts{uri};
        } else {
            $self->{uri} = $config->{rssuri};
        }
        if ($opts{description}) {
            $self->{description} = $opts{description};
        } else {
            $self->{description} = $config->{description};
        }
        if ($opts{xsl}) {
            $self->{xsl} = $opts{xsl};
        } else {
            $self->{xsl} = $config->{xsl};
        }
        if ($self->{parent} !~ m|/$|) {
            $self->{parent} .= "/";
        }
        return bless $self;
        
    }

    sub append($%) {
        my ($self, %opts) = @_;
        my $item = new Item(%opts);
        $self->{items}{$opts{link}} = $item;
    }

    sub keys($) {
        my ($self) = @_;
        my @k = keys %{$self->{items}};
        return reverse(sort(@k));
    }

    sub makeRSS1($) {
        my ($self) = @_;
        my $buf = "";
        $buf .= sprintf('<?xml version="1.0" encoding="%s"?>',
                        $self->{encode}) . "\n";
        if ($self->{xsl}) {
            $buf .= sprintf('<?xml-stylesheet href="%s" type="text/xsl"?>',
                            $self->{xsl}) . "\n";
        }
        $buf .= '<rdf:RDF' . "\n"
           . '  xmlns="http://purl.org/rss/1.0/"' . "\n"
           . '  xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"'."\n"
           . '  xmlns:content="http://purl.org/rss/1.0/modules/content/"'."\n"
           . '  xmlns:dc="http://purl.org/dc/elements/1.1/"' . "\n"
           . '  xml:lang="' . $self->{lang} . '">' . "\n"
           . '<channel rdf:about="' . $self->{uri} . '">' . "\n"
           . '  <title>' . $self->{title} . '</title>' . "\n"
           . '  <link>' . $self->{link} . '</link>' . "\n"
           . '  <description>' . $self->{description} . '</description>'."\n"
           . '  <items><rdf:Seq>' . "\n";
        my $f;
        foreach $f ($self->keys()) {
            $buf .= sprintf('    <rdf:li rdf:resource="%s"/>', $f) . "\n";
        }
        $buf .= '  </rdf:Seq></items>' . "\n"
             . '</channel>' . "\n";

        my $uri;
        foreach $uri ($self->keys()) {
            $f = $self->{items}{$uri};
            $buf .= sprintf('<item rdf:about="%s">', $f->{link}) . "\n"
                 . sprintf('  <title>%s</title>', $f->{title}) . "\n"
                 . sprintf('  <link>%s</link>', $f->{link}) . "\n"
                 . sprintf('  <dc:date>%s</dc:date>', $f->{date}) . "\n";
            if ($f->{description}) {
                  $buf .= '  <description>' . $f->{description}
                       . '</description>' . "\n";
            }
            if ($f->{content}) {
                  $buf .= '  <content:encoded><![CDATA['
                       .  $f->{content} . ']]></content:encoded>' . "\n";
            }
            $buf .= "</item>\n"
        }

        $buf .= "</rdf:RDF>\n";
        return $buf;
    }
#}

package RSSfile; #{
    sub new($$) {
        my ($self, $filename) = @_;
        local $_;
        my $fp;
        open $fp, $filename;
        my @buf = <$fp>;
        close $fp;
        $_ = join("", @buf);
        my @items = ();
        while (m|<item[^<>]*>(.*?)</item>|s) {
            $_ = $';
            my $itemstr = $1;
            push @items, (new Item());
            while (1) {
                if ($itemstr =~ /<title>(.*?)</s) {
                    $items[-1]->{title} = $1;
                    $itemstr = $';
                 } elsif ($itemstr =~ /<link>(.*?)</s) {
                    $items[-1]->{link} = $1;
                    $itemstr = $';
                 } elsif ($itemstr =~ /<dc:date>(.*?)</s) {
                    $items[-1]->{date} = $1;
                    $itemstr = $';
                 } elsif ($itemstr =~ /<description>(.*?)</s) {
                    $items[-1]->{description} = $1;
                    $itemstr = $';
                 } elsif ($itemstr =~ /<content:encoded><!\[CDATA\[(.*?)\]\]>/s) {
                    $items[-1]->{content} = $1;
                    $itemstr = $';
                 } else {
                    last;
                 }
            }
        }
        $self = {items => [@items]};
        return bless $self;
    }

#}

package Entry; #{
    use POSIX qw(strftime);

    sub new($$) {
        my ($self, $id) = @_;
        $self = {id=>$id};
        return bless $self;
    }

    sub escapeHTML($) {
        my ($str) = @_;
        $str =~ s/&/&amp;/g;
        $str =~ s/</&lt;/g;
        $str =~ s/>/&gt;/g;
        return $str;
    }

    sub update($%) {
        my ($self, %opts) = @_;
        my $title = $opts{title} || "";
        my $content = ${opts}{content} || "";
        if (! ($title || $content)) {
            $self->remove();
            return;
        }
        my $tmp = $content;
        $tmp =~ s/<.*?>/ /g;
        my $i;
        my $description = "";
        foreach $i (split(/\s+/, $tmp)) {
            $description .= "$i ";
            if (length($description) >= $config->{descriptionsize}) {
                last;
            }
        }
        $description =~ s/\s+/ /g;
        chop $description;
        my $link = $config->{link} . "#e" . $self->{id};
        my $w3cdate = strftime('%Y-%m-%dT%H:%M:%S', localtime($self->{id}));
        $w3cdate .= $config->{timezone};
        $title = escapeHTML($title);
        $description = escapeHTML($description);
        $content =~ s/]]>/&#93;&#93;>/g;
        $self->{item} = new Item(link => $link,
                                 title => $title,
                                 date => $w3cdate,
                                 description => $description,
                                 content => $content);
        return $self->{item};
    }

    sub month($) {
        my ($self) = @_;
        return strftime("%Y-%m", localtime($self->{id}));
    }

    sub dirname($) {
        my ($self) = @_;
        return $config->{xmldir} . "/" . $self->month();
    }

    sub path($) {
        my ($self) = @_;
        return sprintf("%s/%d.rdf", $self->dirname(), $self->{id});
    }

    sub load($) {
        my ($self) = @_;
        my $path = $self->path();
        if (-f $path) {
            my $rss = new RSSfile($path);
            $self->{item} = $rss->{items}[0];
            return 1;
        } else {
            return 0;
        }
    }

    sub sync($) {
        my ($self) = @_;
        my $dirname = $self->dirname();
        if (! (-d $dirname)) {
            mkdir $dirname, 0777 or
                Error::xdie("MkdirError: $dirname");
        }
        my $rss = new RSS();
        $rss->{items}{$self->{item}{link}} = $self->{item};
        my $fp;
        open $fp, ">". $self->path() or
            Error::xdie("IOError: ".$self->path());
        my $buf = $rss->makeRSS1();
        print $fp $buf;
        close $fp;
        return $buf;
    }

    sub remove($) {
        my ($self) = @_;
        my $path = $self->path();
        my $dirname = $self->dirname();
        unlink $path;
        my @files = glob("$dirname/*");
        if (! @files) {
            rmdir $dirname;
        } elsif (@files == 1) {
            unlink $files[0];
            rmdir $dirname;
        }
    }
#}

package BlogMain; #{
    use Digest::Perl::MD5 qw(md5_hex);

    sub updateArchives() {
        my $rss = new RSS(uri => $config->{archivesuri});
        my $parent = $config->{link};
        my $i;
        foreach $i (glob($config->{xmldir} . "/[0-9]*")) {
            if ($i =~ /[-0-9]+/) {
                $rss->append(link => "$parent#a$&",
                             title => $&,
                             date => $&);
            }
        }
        my $fp;
        open $fp, ">" . $config->{archives};
        print $fp $rss->makeRSS1();
        close $fp;
    }

    sub updateRecent($) {
        my ($rssCache) = @_;
        my $rss = new RSS(uri => $config->{rssuri});
        my @ids = ();
        my $xmldir = $config->{xmldir}; 
        my @month = glob("$xmldir/[0-9]*");
        @month = reverse(sort(@month));
        my $rsssize = $config->{rsssize}; 
        my $i;
        foreach $i (@month) {
            my $j;
            foreach $j (glob("$i/*")) {
                if ($j =~ /(\d+).rdf/) {
                    push @ids, $1;
                }
            }
            if (@ids >= $rsssize) {
                last;
            }
        }
        my $count = 0;
        @ids = sort {$b <=> $a} @ids;
        my $item;
        my $id;
        for $id (@ids) {
            if ($rssCache->{$id}) {
                $item = $rssCache->{$id};
            } else {
                my $entry = new Entry($id);
                $entry->load();
                $item = $entry->{item};
            }
            $rss->{items}{$item->{link}} = $item;
            $rssCache->{$id} = $item;
            $count++;
            if ($count >= $rsssize) {
                last;
            }
        }
        my $fp;
        open $fp, ">" . $config->{recent} or
            Error::xdie("IOError: ". $config->{recent});
        print $fp $rss->makeRSS1();
        close $fp;
    }

    sub updateMonthly($$) {
        my ($month, $rssCache) = @_;
        my $rss = new RSS(uri => $config->{xmluri} . "/$month/index.rdf");
        my $dirname = $config->{xmldir} . "/$month";
        if (! (-d $dirname)) {
            return;
        }
        my $i;
        foreach $i (glob("$dirname/*")) {
            if ($i =~ /(\d+).rdf/) {
                my $id = $1;
                my $item;
                if ($rssCache->{$id}) {
                    $item = $rssCache->{$id};
                } else {
                    my $entry = new Entry($id);
                    $entry->load();
                    $item = $entry->{item};
                }
                $item->{content} = "";
                $rss->{items}{$item->{link}} = $item;
                $rssCache->{$id} = $item;
            }
        }
        my $fp;
        open $fp, "> $dirname/index.rdf" or
            Error::xdie("IOError: $dirname/index.rdf");
        print $fp $rss->makeRSS1();
        close $fp;
    }

    sub writeXML($$$) {
        my ($id, $title, $body) = @_;
        $title =~ s/(^\s*|\s*$)//g;
        $body =~ s/(^\s*|\s*$)//g;
        if (! $id) {
            $id = time();
        } elsif ($id !~ /^\d+$/) {
            Error::error("Wrong entry ID.");
            return;
        }
        if (length($title) + length($body) > $config->{pagesize}) {
            Error::error("Too large entry.");
            return;
        }
        my $entry = new Entry($id);
        my $item = $entry->update(title => $title, content => $body);
        my $rssCache = {};
        $rssCache->{$id} = $item;
        if ($title || $body) {
            $entry->sync();
        }
        updateArchives();
        updateRecent($rssCache);
        updateMonthly($entry->month(), $rssCache);
        if ($title || $body) {
            print "Content-Type: application/xml;\n";
            print "\n";
            print '<?xml version="1.0" encoding="utf-8"?>', "\n";
            print '<ok>ok</ok>', "\n";
        } else {
            Error::error("Removed");
        }
    }

    sub main() {
        $config = new BlogConfig("blog.conf");
        $ENV{TZ} = $config->{tz};
        my $q = new CGI();
        my $id = $q->param("id") || "";
        my $passwd = $q->param("passwd") || "";
        my $body = $q->param("body") || "";
        $body =~ /\n/;
        my $title = $`;
        $body = $';
        if (md5_hex($passwd) eq $config->{passwd}) {
            writeXML($id, $title, $body);
        } else {
            Error::error("Wrong password.");
        }

    }
#}

BlogMain::main();
exit;

__END__
